@@ -2,7 +2,7 @@
 
 /* { dg-options "-pedantic" } */
 
-void foo(unsigned long* ulp);
+void foo(unsigned long* ulp);/* { dg-message "note: expected '\[^'\n\]*' but argument is of type '\[^'\n\]*'" "note: expected" { target *-*-* } 5 } */
 
 void bar(long* lp) {
   foo(lp); /* { dg-warning "differ in signedness" } */
