@@ -1,9 +1,9 @@
-// Copyright (C) 2006 Free Software Foundation, Inc.
+// Copyright (C) 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
 // terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2, or (at your option)
+// Free Software Foundation; either version 3, or (at your option)
 // any later version.
 
 // This library is distributed in the hope that it will be useful,
@@ -12,9 +12,8 @@
 // GNU General Public License for more details.
 
 // You should have received a copy of the GNU General Public License along
-// with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
-// USA.
+// with this library; see the file COPYING3.  If not see
+// <http://www.gnu.org/licenses/>.
 
 // 23.2.2.2 list capacity [lib.list.capacity]
 
@@ -28,7 +27,13 @@ void test01()
 
   std::list<int> l;
 
-  VERIFY( l.max_size() == l.get_allocator().max_size() );
+#ifndef _GLIBCXX_DEBUG
+  using std::_List_node;
+#else
+  using std::_GLIBCXX_STD_D::_List_node;
+#endif
+
+  VERIFY( l.max_size() == std::allocator<_List_node<int> >().max_size() );
 }
 
 int main()
