@@ -1,9 +1,9 @@
-// Copyright (C) 2003 Free Software Foundation, Inc.
+// Copyright (C) 2003, 2009 Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
 // terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2, or (at your option)
+// Free Software Foundation; either version 3, or (at your option)
 // any later version.
 
 // This library is distributed in the hope that it will be useful,
@@ -12,9 +12,8 @@
 // GNU General Public License for more details.
 
 // You should have received a copy of the GNU General Public License along
-// with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
-// USA.
+// with this library; see the file COPYING3.  If not see
+// <http://www.gnu.org/licenses/>.
 
 // { dg-require-fileio "" }
 
@@ -47,12 +46,12 @@ protected:
   {
     unshift_called = true;
     to_next = to;
-    return std::codecvt_base::ok;
+    return codecvt_base::error;
   }
 };
 
 // libstdc++/12790
-// basic_filebuf::close() should call codecvt::unshift()
+// basic_filebuf::close() should fail if codecvt::unshift() fails
 void test01()
 {
   using namespace std;
@@ -68,11 +67,11 @@ void test01()
 
   fb.open(name, ios_base::out);
   fb.sputc('a');
-  fb.pubsync(); // Does not call unshift()
 
   VERIFY( !cvt->unshift_called );
-  fb.close();
+  filebuf* ret = fb.close();
   VERIFY( cvt->unshift_called );
+  VERIFY( ret == NULL );
 }
 
 int main()
