@@ -1,11 +1,12 @@
 // 2004-10-20  Benjamin Kosnik  <bkoz@redhat.com>
 //
-// Copyright (C) 2004 Free Software Foundation, Inc.
+// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009
+// Free Software Foundation, Inc.
 //
 // This file is part of the GNU ISO C++ Library.  This library is free
 // software; you can redistribute it and/or modify it under the
 // terms of the GNU General Public License as published by the
-// Free Software Foundation; either version 2, or (at your option)
+// Free Software Foundation; either version 3, or (at your option)
 // any later version.
 //
 // This library is distributed in the hope that it will be useful,
@@ -14,9 +15,8 @@
 // GNU General Public License for more details.
 //
 // You should have received a copy of the GNU General Public License along
-// with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
-// USA.
+// with this library; see the file COPYING3.  If not see
+// <http://www.gnu.org/licenses/>.
 
 // 6.2.2 Class template array
 
@@ -29,10 +29,10 @@ test01()
   const size_t len = 5;
   typedef std::tr1::array<int, len> array_type;
   bool test __attribute__((unused)) = true;
-  array_type a = { 0, 1, 2, 3, 4 };
-  array_type b = { 0, 1, 2, 3, 4 };
-  array_type c = { 0, 1, 2, 3 };
-  
+  array_type a = { { 0, 1, 2, 3, 4 } };
+  array_type b = { { 0, 1, 2, 3, 4 } };
+  array_type c = { { 0, 1, 2, 3 } };
+
   VERIFY( !(a != b) );
   VERIFY( a != c );
 }
